-------------------------------------------------------------------------------
-- Title      : EMIF/buffer memory interface
-- Project    : Graphic LCD controller
-------------------------------------------------------------------------------
-- File       : dispbufctrl.vhdl
-- Author     : Andy Peters  <devel@latke.net>
-- Company    : ASP Digital
-- Created    : 2006-07-28
-- Last update: 2006-08-12
-- Platform   : ModelSim, XST
-- Standard   : VHDL'02, Math Packages
-------------------------------------------------------------------------------
-- Description: This module interfaces the C8051F340 microcontroller to the LCD
-- controller display buffer.
-- We assume that the '340's EMIF signals are synchronous to its SYSCLK output.
-- The LCD controller sees the display buffer as a 4-bit by 16384 (we only need
-- 15360) word memory.  Of course, since the '340 EMIF is 8 bits wide, we use a
-- dual-port memory with different-size ports.  This memory is instantiated
-- from a Xilinx Coregen-created component.
--
-- EMIF timing: Address setup = 1 SYSCLK
--              Pulse width = 1 SYSCLK
--              Address hold = 2 SYSCLKs
--
-- A note about the dual-port RAM:
-- The "B" side is accessed by the micro.  Its data port is 8 bits wide, and
-- the address bus is 13 bits wide.  The B side has a write enable.
-- The "A" side is accessed by the glcd display controller to read data.  Its
-- data port is 4 bits wide and its addrses bus is 14 bits wide.  This port is
-- read only. 
-------------------------------------------------------------------------------
-- Copyright (c) 2006 ASP Digital
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2006-07-28  1.0      andy    Created
-- 2006-07-29           andy    Decode 8051 address and gate write strobe with
--                              decode result.
-- 2006-08-03           andy    display buffer memory is now 8 bits wide on
--                              '340 side and one bit wide on display
--                              controller side.  Use four dual ports, one for
--                              each quadrant.
--                              Display buffer memory component named changed
--                              to dispbuf from dpram.
--                              Add reset input.
-- 2006-08-09           andy    reverse mux decode bits! idecode(0) is the
--                              left-most bit!  ARRRRRGH!
-- 2006-08-12           andy    add EMIF timing note in comments.
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity dispbufctrl is
    generic (
        DISPMEMBASE : unsigned(15 downto 0) := X"1000";  -- start of display buf
        DISPMEMSIZE : unsigned(15 downto 0) := X"0800");  -- size of disp buffer
    port (
        SysClk    : in  std_logic;      -- global SYSCLK from '340
        Reset_l   : in  std_logic;      -- global reset
        EmifA     : in  std_logic_vector(15 downto 0);   -- '340 ext address
        EmifDIn   : in  std_logic_vector(7 downto 0);    -- data from '340
        EmifDOut  : out std_logic_vector(7 downto 0);    -- data out to '340
        EmifWr_l  : in  std_logic;      -- '340 ext write strobe
        ClkPix4x  : in  std_logic;      -- clock 4x the pixel clock
        DispData  : out std_logic_vector(0 to 3);        -- display data
        DispDAddr : in  std_logic_vector(13 downto 0));  -- disp buffer addr
end entity dispbufctrl;

architecture dualport of dispbufctrl is

    component dispbuf
        port (
            addra : in  std_logic_vector(13 downto 0);
            addrb : in  std_logic_vector(10 downto 0);
            clka  : in  std_logic;
            clkb  : in  std_logic;
            dinb  : in  std_logic_vector(7 downto 0);
            douta : out std_logic_vector(0 downto 0);
            doutb : out std_logic_vector(7 downto 0);
            web   : in  std_logic);
    end component dispbuf;

    signal iweb      : std_logic_vector(0 to 3);  -- RAM write enables
    signal iDecode   : std_logic_vector(0 to 3);  -- quadrant decodes
    type   arr4slv8 is array (0 to 3) of std_logic_vector(7 downto 0);
    signal iEmifDOut : arr4slv8;                  -- each memory's output bus
    
begin  -- architecture dualport

    -- Decode incoming address.  If we're in the right "Space" to access the
    -- display buffer memory, we can write.
    memgen : for m in 0 to 3 generate
        Decoder : process (SysClk, Reset_l) is
        begin  -- process Decoder
            if Reset_l = '0' then           -- asynchronous reset (active low)
                iDecode(m) <= '0';
            elsif rising_edge(SysClk) then  -- rising clock edge
                if ((unsigned(EmifA) >= DISPMEMBASE + (m * DISPMEMSIZE)) and
                    (unsigned(EmifA) < (DISPMEMBASE + ((m + 1) * DISPMEMSIZE)))) then
                    iDecode(m) <= '1';
                else
                    iDecode(m) <= '0';
                end if;
            end if;
        end process Decoder;

        -- decode the four quadrant address spaces and generate a write enable:
        iweb(m) <= '1' when ((EmifWr_l = '0') and (iDecode(m) = '1'))
                   else '0';
        
        u_dispbuf : dispbuf
            port map (
                addra => DispDAddr,
                addrb => EmifA(10 downto 0),
                clka  => ClkPix4x,
                clkb  => SysClk,
                dinb  => EmifDIn,
                douta => DispData(m to m),  -- single bit vector.
                doutb => iEmifDOut(m),
                web   => iweb(m));
    end generate memgen;

    -- EmifDOut mux:
    EmifDOutMux : process (SysClk, Reset_l) is
    begin  -- process EmifDOutMux
        if Reset_l = '0' then           -- asynchronous reset (active low)
            EmifDOut <= (others => '0');
        elsif rising_edge(SysClk) then  -- rising clock edge
            Decoder : case iDecode is
                when "1000" => EmifDOut <= iEmifDOut(0);
                when "0100" => EmifDOut <= iEmifDOut(1);
                when "0010" => EmifDOut <= iEmifDOut(2);
                when "0001" => EmifDOut <= iEmifDOut(3);
                when others => EmifDOut <= X"AA";
            end case Decoder;
        end if;
    end process EmifDOutMux;

end architecture dualport;
