-------------------------------------------------------------------------------
-- Title      : Graphic LCD (LM215XB) Controller
-- Project    : Graphic LCD (LM215XB) Controller
-------------------------------------------------------------------------------
-- File       : glcd.vhdl
-- Author     : Andy Peters  <devel@latke.net>
-- Company    : ASP Digital
-- Created    : 2006-07-27
-- Last update: 2006-08-06
-------------------------------------------------------------------------------
-- Description: This is a controller for the LM215XB 480x128 Graphic LCD.
-- This LCD has four quadrants, each 240 pixels x 64 lines.
-- There are four serial data lines (one for each quadrant) and four control
-- signals.  The control signals are:
-- CL1: Basically, a horizontal sync, asserted at the start of a line for at
-- least 150 ms.
-- CL2: Pixel clock, data is set up to the falling edge.
-- FLM: Basically, a vertical sync, apparently asserted for a line time on the
-- first line.
-- M: AC signal that toggles every field.
-- D1: upper-left quadrant data input.
-- D2: lower-left quadrant data input.
-- D3: upper-right quadrant data input.
-- D4: lower-right quadrant data input.
--
-- This module assumes that there is a display buffer using four one-bit-wide
-- Xilinx block RAMs, one for each quadrant.  Some other process is responsible
-- for filling the display buffer.
--
-- This module runs off of a clock that's four times the pixel clock.  With a
-- 48 MHz global oscillator, divided by 10 using the DCM, this module's input
-- clock is 4.8 MHz, which we divide here to get a CL2 frequency of 1.2 MHz,
-- right near the LM215XB's recommended 1.15 MHz spec.
-------------------------------------------------------------------------------
-- Copyright (c) 2006 Spectral Instruments, Inc.
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2006-07-27  1.0      andy    Created
-- 2006-07-28           andy    display buffer address bus is 14 bits (16k)
-- 2006-08-03           andy    use four independent one-bit-wide dual ports,
--                              one for each quadrant.  Quadrants are numbered
--                              0 to 3, not 1 to 4.
-- 2006-08-06           andy    modify FLM, M timing to hopefully match data
--                              sheet.
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity glcd is
    generic (
        PIXPERLINE    : integer := 240;             -- # of pixels in a line
        LINESPERFIELD : integer := 64);             -- # of lines in a field
    port (
        Clk4x     : in  std_logic;      -- Clock 4x the pixel clock
        Reset_l   : in  std_logic;      -- global reset
        DispData  : in  std_logic_vector(0 to 3);   -- display data
        DispDAddr : out std_logic_vector(13 downto 0);  -- disp buffer address
        GlcdCL1   : out std_logic;      -- LCD horizontal sync
        GlcdCL2   : out std_logic;      -- LCD pixel clock
        GlcdFLM   : out std_logic;      -- LCD vertical sync
        GlcdM     : out std_logic;      -- LCD AC strobe
        GlcdD     : out std_logic_vector(0 to 3));  -- LCD data for each quad
end entity glcd;

architecture controller of glcd is
    -- The graphic LCD has some clocks and syncs that we need to generate.
    -- We get the pixel clock (CL2) by dividing our input clock by 4:
    signal iClk4xCnt : unsigned(1 downto 0);

    -- count pixels, so we can generate the horizontal sync (CL1).
    signal iPixCnt : natural range 0 to PIXPERLINE-1;

    -- count lines, so we can generate the vertical sync (FLM).
    signal iLineCnt : natural range 0 to LINESPERFIELD-1;

    -- internal FLM, use for edge detect:
    signal iGlcdFLM : std_logic;

    -- internal GlcdM, use for toggle every field:
    signal iGlcdM : std_logic;

    -- memory address counter:
    signal iDispDAddr : unsigned(13 downto 0);

begin  -- architecture controller

    -- Generate the pixel clock, and count the pixels.
    PixClkGen : process (Clk4x, Reset_l) is
    begin  -- process PixClkGen
        if Reset_l = '0' then
            iClk4xCnt <= (others => '0');
            GlcdCL2   <= '0';
        elsif rising_edge(Clk4x) then
            iClk4xCnt <= (iClk4xCnt + 1) mod 4;
            GlcdCL2   <= not iClk4xCnt(1);
        end if;
    end process PixClkGen;

    -- count pixels:
    CountPixels : process (Clk4x, Reset_l) is
    begin  -- process CountPixels
        if Reset_l = '0' then
            iPixCnt <= 208;
        elsif rising_edge(Clk4x) then
            if (iClk4xCnt = 3) then
                if (iPixCnt = (PIXPERLINE-1)) then
                    iPixCnt <= 0;
                else
                    iPixCnt <= iPixCnt + 1;
                end if;
            end if;
        end if;
    end process CountPixels;

    -- drive the horizontal sync.
    DriveGlcdCL1 : process (Clk4x, Reset_l) is
    begin  -- process DriveGlcdCL1
        if Reset_l = '0' then
            GlcdCL1 <= '0';
        elsif rising_edge(Clk4x) then
            if ((iPixCnt = (PIXPERLINE-1)) and
                ((iClk4xCnt = 2) or (iClk4xCnt = 3))) then
                GlcdCL1 <= '1';
            else
                GlcdCL1 <= '0';
            end if;
        end if;
    end process DriveGlcdCL1;

    -- count lines:
    CountLines : process (Clk4x, Reset_l) is
    begin  -- process CountLines
        if Reset_l = '0' then
            iLineCnt <= 62;
        elsif rising_edge(Clk4x) then
            if (iClk4xCnt = 3) and (iPixCnt = (PIXPERLINE-1)) then
                if (iLineCnt = (LINESPERFIELD-1)) then
                    iLineCnt <= 0;
                else
                    iLineCnt <= iLineCnt + 1;
                end if;
            end if;
        end if;
    end process CountLines;

    -- drive vertical sync on the first line of the field.
    -- data sheet sez FLM should be asserted at least 2 us after the dropping
    -- edge of CL1, and it has to go away between 0 and 100 ns after the
    -- dropping edge of the next CL1.
    -- So: for the assertion, wait three pixel clocks (iPixCnt = 2) before
    -- asserting FLM.
    DriveFLM : process (Clk4x, Reset_l) is
    begin  -- process DriveFLM
        if Reset_l = '0' then
            iGlcdFLM <= '0';
        elsif rising_edge(Clk4x) then
            if iLineCnt = 1 then
                iGlcdFLM <= '0';
            elsif ((iLineCnt = 0) and (iPixCnt = 3)) then
                iGlcdFLM <= '1';
            end if;
            GlcdFLM <= iGlcdFLM;

        end if;
    end process DriveFLM;

    -- toggle M on the falling edge of FLM:
    DriveM : process (Clk4x, Reset_l) is
    begin  -- process DriveM
        if Reset_l = '0' then
            iGlcdM <= '0';
            GlcdM  <= '0';
        elsif rising_edge(Clk4x) then
            if (iLineCnt = 1) and (iGlcdFLM = '1') then
                iGlcdM <= not iGlcdM;
            end if;
            GlcdM <= iGlcdM;
        end if;
    end process DriveM;


    -- Handle display buffer access:
    BufAddr : process (Clk4x, Reset_l) is
    begin  -- process BufAddr
        if Reset_l = '0' then
            iDispDAddr <= (others => '0');
        elsif falling_edge(Clk4x) then
            if ((iPixCnt = 0) and (iLineCnt = 0)) then
                iDispDAddr <= (others => '0');
            elsif (iClk4xCnt = 0) then
                iDispDAddr <= iDispDAddr + 1;
            end if;
        end if;
    end process BufAddr;

    DispDAddr <= std_logic_vector(iDispDAddr);

    DelayData : process (Clk4x, Reset_l) is
    begin  -- process DelayData
        if Reset_l = '0' then           -- asynchronous reset (active low)
            GlcdD <= (others => '0');
        elsif rising_edge(Clk4x) then   -- rising clock edge
            GlcdD <= DispData;
        end if;
    end process DelayData;

end architecture controller;
