-------------------------------------------------------------------------------
-- Title      : Top-level test module for a graphic LCD interface
-- Project    : 
-------------------------------------------------------------------------------
-- File       : glcdtest.vhdl
-- Author     : Andy Peters  <devel@latke.net>
-- Company    : ASP Digital
-- Created    : 2006-07-27
-- Last update: 2006-08-12
-- Platform   : ModelSim, XST
-- Standard   : VHDL'02, Math Packages
-------------------------------------------------------------------------------
-- Description: This chip design interfaces a SiLabs C8051F340 to an LM215XB
-- 480-by-128 graphic LCD panel.  The FPGA design is fairly dumb, in the sense
-- that the 8051 is responsible for loading the display buffer.  All we do here
-- is generate the LCD control signals while reading from the display buffer.
--
-- We interface to the 8051 using its EMIF in the 16-bit non-multiplexed mode.
-- This interface uses P2 for the upper address byte, P3 for the lower address
-- byte and P4 for data in/out.  Also needed are the read and write strobes,
-- although readback isn't really all that interesting. We assume that the EMIF
-- is synchronous to the 8051's SYSCLK output, so we use that clock for this
-- interface.
--
-- EMIF timing: Address setup = 1 SYSCLK
--              Pulse width = 1 SYSCLK
--              Address hold = 2 SYSCLKs
--
-- The interface to the LCD consists of four data lines (one for each
-- quadrant), a pixel clock (CL2), a horizontal sync strobe (CL1), a vertical
-- sync strobe (FLM) and an AC toggle (M).  The global 48 MHz clock is divided
-- down by 10 to 4.8 MHz using a DCM to provide a 4x clock to the GLCD
-- controller module.  This 4x clock is divided down to provide the 1.2 MHz
-- pixel clock to the LCD.
-- 
-------------------------------------------------------------------------------
-- Copyright (c) 2006 ASP Digital
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2006-07-27  1.0      andy    Created
-- 2006-07-28           andy    Display buffer memory address bus is actually
--                              14 bits wide (to address 16k words)
-- 2006-07-29           andy    dispbufctrl needs entire 8051 address to
--                              decode. 
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

entity glcdtest is
    generic (
        PIXPERLINE    : integer := 240;  -- # of pixels in a line
        LINESPERFIELD : integer := 64);  -- # of lines in a field
    port (
        SysClk   : in    std_logic;     -- 48 MHz SYSCLK from '340
        Reset_l  : in    std_logic;     -- global reset from '340
        EmifA    : in    std_logic_vector(15 downto 0);  -- '340 ext address
        EmifD    : inout std_logic_vector(7 downto 0);   -- '340 ext data
        EmifWr_l : in    std_logic;     -- '340 ext write strobe
        EmifRd_l : in    std_logic;     -- '340 ext read strobe
        Ps2Clk   : in    std_logic;     -- PS/2 I/F clock in
        Ps2Data  : in    std_logic;     -- PS/2 I/F data input
        SpiMOSI  : out   std_logic;     -- SPI Data to micro
        SpiMISO  : in    std_logic;     -- SPI data from micro
        SpiClk   : out   std_logic;     -- SPI interface clock to micro
        SpiNSS_l : out   std_logic;     -- SPI slave select
        GlcdCl1  : out   std_logic;     -- LCD horiz sync
        GlcdCL2  : out   std_logic;     -- LCD pixel clock
        GlcdFLM  : out   std_logic;     -- LCD vertical sync
        GlcdM    : out   std_logic;     -- LCD AC toggle
        GlcdD    : out   std_logic_vector(0 to 3));      -- LCD data lines
end entity glcdtest;

architecture toplevel of glcdtest is
    -- DCM likes an active-high reset:
    signal Reset : std_logic;

    -- from DCM:
    signal iSysClk   : std_logic;       -- buffered SysClk
    signal iClkPix4x : std_logic;       -- clock 4x the pixel clock

    -- from dualport's display-buffer side:
    signal iDispData  : std_logic_vector(0 to 3);
    signal iDispDAddr : std_logic_vector(13 downto 0);

    -- handle bidirectional '340 data bus:
    signal iEmifRd_l_d : std_logic_vector(2 downto 0);  -- delay EMIF read
    signal iEmifDIn    : std_logic_vector(7 downto 0);  -- register data
    signal iEmifDOut   : std_logic_vector(7 downto 0);  -- data TO '340

    -- PS/2 interface:
    signal iScanCode      : std_logic_vector(7 downto 0);  -- keypress
    signal iScanCodeValid : std_logic;  -- true when new keycode ready
    signal iScanCodeRd    : std_logic;  -- "read" scan code

    constant GND : std_logic := '0';

    component gdcm
        port (
            CLKIN_IN        : in  std_logic;
            RST_IN          : in  std_logic;
            CLKDV_OUT       : out std_logic := '0';
            CLKIN_IBUFG_OUT : out std_logic := '0';
            CLK0_OUT        : out std_logic := '0';
            CLKFX_OUT       : out std_logic := '0');
    end component gdcm;

    component BUFG is
        port (
            O : out std_ulogic;
            I : in  std_ulogic);
    end component BUFG;

begin  -- architecture toplevel

    Reset <= not Reset_l;

    u_gdcm : gdcm
        port map (
            CLKIN_IN        => SysClk,
            RST_IN          => Reset,
            CLKDV_OUT       => iClkPix4x,
            CLKIN_IBUFG_OUT => open,
            CLK0_OUT        => iSysClk,
            CLKFX_OUT       => SpiClk);

    u_glcd : entity work.glcd
        generic map (
            PIXPERLINE    => PIXPERLINE,
            LINESPERFIELD => LINESPERFIELD)
        port map (
            Clk4x     => iClkPix4x,     -- clock 4x the pixel clock
            Reset_l   => Reset_l,       -- global reset
            DispData  => iDispData,     -- display buffer data
            DispDAddr => iDispDAddr,    -- disp buffer address
            GlcdCL1   => GlcdCL1,       -- LCD horizontal sync
            GlcdCL2   => GlcdCL2,       -- LCD pixel clock
            GlcdFLM   => GlcdFLM,       -- LCD vertical sync
            GlcdM     => GlcdM,         -- LCD AC strobe
            GlcdD     => GlcdD);        -- LCD data for each quadrant

    SyncDataIn : process (iSysClk, Reset_l) is
    begin  -- process SyncDataIn
        if Reset_l = '0' then
            iEmifDIn <= (others => '0');
        elsif rising_edge(iSysClk) then
            iEmifDIn <= EmifD;
        end if;
    end process SyncDataIn;
    u_dispbufctrl : entity work.dispbufctrl
        port map (
            SysClk    => iSysClk,       -- global SYSCLK from '340
            Reset_l   => Reset_l,       -- global reset
            EmifA     => EmifA,         -- '340 ext address
            EmifDIn   => iEmifDIn,      -- data from '340
            EmifDOut  => iEmifDOut,     -- data out to '340
            EmifWr_l  => EmifWr_l,      -- '340 ext write strobe
            ClkPix4x  => iClkPix4x,     -- clock 4x the pixel clock
            DispData  => iDispData,     -- display data
            DispDAddr => iDispDAddr);   -- disp buffer addr

    -- Delay EmifRd_l to ensure that read data remains valid after EmifRd_l
    -- goes away.
    DelayEmifRd : process (iSysClk) is
    begin  -- process DelayEmifRd
        if rising_edge(iSysClk) then
            iEmifRd_l_d <= iEmifRd_l_d(1 downto 0) & EmifRd_l;
        end if;
    end process DelayEmifRd;
    -- tristate on '340 data bus:    
    EmifD <= iEmifDOut when ((iEmifRd_l_d(2) = '0') or
                             (iEmifRd_l_d(1) = '0') or
                             (iEmifRd_l_d(0) = '0') or
                             (EmifRd_l = '0')) else
             (others => 'Z');

    u_ps2kbdif : entity work.ps2kbdif
        port map (
            SysClk        => iClkPix4x,        -- global clock
            Reset_l       => Reset_l,          -- global reset
            Ps2clk        => Ps2Clk,           -- PS/2 I/F clock
            Ps2Data       => Ps2Data,          -- PS/2 I/F data
            ScanCode      => iScanCode,        -- scan code from keyboard
            ScanCodeRd    => iScanCodeRd,      -- read scan code
            ScanCodeValid => iScanCodeValid);  -- true when scan valid

    -- "read" the keyboard interface by simply using the ScanCodeValid signal
    -- delayed by a clock as the read.  Also use it as the SPI parallel-data
    -- write, since the data is valid before the Valid flag is asserted.
    WriteScanCode : process (iClkPix4x, Reset_l) is
    begin  -- process WriteScanCode
        if Reset_l = '0' then
            iScanCodeRd <= '0';
        elsif rising_edge(iClkPix4x) then
            iScanCodeRd <= iScanCodeValid;
        end if;
    end process WriteScanCode;

    u_spimaster : entity work.spimaster
        port map (
            SpiClk       => iClkPix4x,    -- SPI clock
            Reset_l      => Reset_l,      -- master reset
            ParDataOut   => iScanCode,    -- data going out
            ParDataWr    => iScanCodeRd,  -- load something into out shifter
            ParDataIn    => open,         -- data coming in
            ParDataInVal => open,         -- true when new rx data available
            ParDataRd    => GND,          -- acknowledge the new data
            SpiMOSI      => SpiMOSI,      -- serial data going out
            SpiMISO      => SpiMISO,      -- serial data coming in
            SpiNSS_l     => SpiNSS_l);    -- slave select out

end architecture toplevel;
