-------------------------------------------------------------------------------
-- Title      : PS/2 Keyboard Scanner
-- Project    : 
-------------------------------------------------------------------------------
-- File       : ps2kbdif.vhdl
-- Author     : Andy Peters  <devel@latke.net>
-- Company    : ASP Digital
-- Created    : 2006-07-30
-- Last update: 2006-07-30
-- Platform   : ModelSim, XST
-- Standard   : VHDL'93, Math Packages
-------------------------------------------------------------------------------
-- Description: This module monitors the PS/2 clock and data signals and
-- returns the scan code of the last key pressed.  The keyboard sends data when
-- it feels like it using a synchronous serial port.  The keyboard writes new
-- bits on the wire when the clock line is high, and the host (in this case,
-- the FPGA) can read it when the clock line is low.
-- The serial format is simple.  The format is a stop bit at logic low,
-- followed by eight data bits (LSB first), followed by an odd parity bit, and
-- terminated with a high stop bit.
--
-- When a code has shifted in and the parity checks OK, the code is latched and
-- the new-code-available flag is asserted.  The code read strobe clears the
-- flag. 
-------------------------------------------------------------------------------
-- Copyright (c) 2006 ASP Digital
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2006-07-30  1.0      andy    Created
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

entity ps2kbdif is
    generic (
        SYNCSIZE : natural := 7);       -- how many sync flops
    port (
        SysClk        : in  std_logic;  -- system clock
        Reset_l       : in  std_logic;  -- global reset
        Ps2clk        : in  std_logic;  -- PS/2 clock
        Ps2Data       : in  std_logic;  -- PS/2 serial data
        ScanCode      : out std_logic_vector(7 downto 0);  -- rx'd scan code
        ScanCodeRd    : in  std_logic;  -- read scan code (clear flag)
        ScanCodeValid : out std_logic);  -- true when code shifted in
end entity ps2kbdif;

architecture shifter of ps2kbdif is

    -- synchronizing flip-flops:
    signal iPs2Clk_d  : std_logic_vector(SYNCSIZE downto 0);
    signal iPs2Data_d : std_logic_vector(SYNCSIZE downto 0);

    -- input shift register:
    signal iShiftReg : std_logic_vector(7 downto 0);  -- shift in here
    signal iParCalc  : std_logic;       -- on-the-fly parity
    signal iParity   : std_logic;       -- received parity
    signal iDone     : std_logic;       -- indicates state machine done

    -- bit counter:
    signal iBitCnt : natural range 0 to 7;

    -- State machine:
    type states_tp is (
        S_RESET,
        S_INIT,
        S_IDLE,
        S_WAIT,
        S_SHIFT,
        S_PARITY,
        S_STOP);
    signal iState : states_tp;
    
    
begin  -- architecture shifter

    -- Sync incoming data and clock.  No reset allows the use of an SRL.
    Sync : process (SysClk) is
    begin  -- process Sync
        if rising_edge(SysClk) then
            iPs2Data_d <= iPs2Data_d(SYNCSIZE-1 downto 0) & Ps2Data;
            iPs2Clk_d  <= iPs2Clk_d(SYNCSIZE-1 downto 0) & Ps2Clk;
        end if;
    end process Sync;

    -- State machine.
    StateMachine : process (SysClk, Reset_l) is
    begin  -- process StateMachine
        if Reset_l = '0' then
            iShiftReg <= (others => '1');
            iParCalc  <= '0';
            iBitCnt   <= 7;
            iParity   <= '0';
            iDone     <= '0';
            iState    <= S_RESET;
        elsif rising_edge(SysClk) then
            case iState is
                when S_RESET =>
                    iShiftReg <= (others => '1');
                    iParCalc  <= '0';
                    iBitCnt   <= 7;
                    iParity   <= '0';
                    iDone     <= '0';
                    iState    <= S_INIT;

                when S_INIT =>
                    -- ensure that the PS/2 lines are both high before continuing:
                    if ((iPs2Clk_d(SYNCSIZE) = '1') and
                        (iPs2Data_d(SYNCSIZE) = '1')) then
                        iState <= S_IDLE;
                    end if;
                    
                when S_IDLE =>
                    iParCalc <= '0';
                    iBitCnt  <= 7;
                    iDone    <= '0';
                    -- Wait for a start bit, which occurs when both clock and
                    -- data are low:
                    if ((iPs2Clk_d(SYNCSIZE) = '0') and
                        (iPs2Data_d(SYNCSIZE) = '0')) then
                        iState <= S_WAIT;
                    end if;

                when S_WAIT =>
                    -- wait for next falling edge of the clock, which is the
                    -- start of the next data bit:
                    if ((iPs2Clk_d(SYNCSIZE) = '1') and
                        (iPs2Clk_d(SYNCSIZE-1) = '0')) then
                        iState <= S_SHIFT;
                    end if;

                when S_SHIFT =>
                    -- by the time the clock edge gets to the end of the
                    -- sync shift register, the data at the start of the SR
                    -- is valid.  Shift in new data, LSB first.
                    iShiftReg <= iPs2Data_d(1) & iShiftReg(7 downto 1);

                    -- update the parity calculation:
                    iParCalc <= iParCalc xor iPs2Data_d(1);

                    -- if we just shifted in the last bit, look for the
                    -- stop bit, otherwise keep waiting and shifting.
                    if iBitCnt = 0 then
                        iState <= S_PARITY;
                    else
                        istate  <= S_WAIT;
                        iBitCnt <= iBitCnt - 1;
                    end if;

                when S_PARITY =>
                    -- The next bit to shift in is the parity bit.  When we get
                    -- it, test it against our calculated parity.
                    if ((iPs2Clk_d(SYNCSIZE) = '1') and
                        (iPs2Clk_d(SYNCSIZE-1) = '0')) then
                        iParity <= iPs2Data_d(1);
                        iState  <= S_STOP;
                    end if;

                when S_STOP =>
                    -- wait for the stop bit before finishing up:
                    if ((iPs2Clk_d(SYNCSIZE) = '1') and
                        (iPs2Clk_d(SYNCSIZE-1) = '0')) then
                        iDone  <= '1';
                        iState <= S_IDLE;
                    end if;

                when others => null;
            end case;
        end if;
    end process StateMachine;

    -- when state machine has shifted everything in, it'll assert Done.  That's
    -- when we test parity and tell the world a new byte is ready.
    -- Done is asserted for exactly one clock tick.
    TestParity : process (SysClk, Reset_l) is
    begin  -- process TestParity
        if Reset_l = '0' then
            ScanCodeValid <= '0';
            ScanCode      <= (others => '0');
        elsif rising_edge(SysClk) then
            if ScanCodeRd = '1' then
                ScanCodeValid <= '0';
            elsif iDone = '1' then
                if iParity = not iParCalc then  -- note odd parity!
                    ScanCode      <= iShiftReg;
                    ScanCodeValid <= '1';
                end if;
            end if;
        end if;
    end process TestParity;
    
end architecture shifter;
