-------------------------------------------------------------------------------
-- Title      : SPI master interface
-- Project    : 
-------------------------------------------------------------------------------
-- File       : spimaster.vhdl
-- Author     : Andy Peters  <devel@latke.net>
-- Company    : ASP Digital
-- Created    : 2006-07-29
-- Last update: 2006-07-30
-- Platform   : ModelSim, XST
-- Standard   : VHDL'93, Math Packages
-------------------------------------------------------------------------------
-- Description: This is a simple SPI master interface.
-- We shift eight-bit words.  The outgoing data bits (MOSI) change on the
-- falling edge of the clock, and we shift incoming bits (MISO) in on the
-- falling edge of the clock.  We assert the slave select on the falling edge
-- of the clock, coincident with the first data bit shifting out.
--
-- The SPI clock runs continually.  Slave select is asserted to let the slave
-- know that it should shift in what's on MOSI.
-- 
-- A new word can be loaded into the outgoing shift register any time the slave
-- select is deasserted (slave select is basically a busy flag).  This is done
-- by asserting ParDataWr.  The outgoing data word is written directly to the
-- shift register.
--
-- When a new word has been received, ParDataInVal will be asserted.  "Read"
-- the new word by asserting ParDataRd, which simply clears the ParDataInVal
-- flag.  Basically, this means that if one wishes to read from the SPI slave,
-- we first have to send something to say, "send me something," and then do
-- another write to that slave, which shifts in the slave's response.
--
-- Of course SPI shifts out MSB first.
-------------------------------------------------------------------------------
-- Copyright (c) 2006 ASP Digital
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2006-07-29  1.0      andy    Created
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

entity spimaster is
    port (
        SpiClk       : in  std_logic;   -- SPI clock
        Reset_l      : in  std_logic;   -- master reset
        ParDataOut   : in  std_logic_vector(7 downto 0);  -- data going out
        ParDataWr    : in  std_logic;   -- load something into out shifter
        ParDataIn    : out std_logic_vector(7 downto 0);  -- data coming in
        ParDataInVal : out std_logic;   -- true when new rx data available
        ParDataRd    : in  std_logic;   -- acknowledge the new data
        SpiMOSI      : out std_logic;   -- serial data going out
        SpiMISO      : in  std_logic;   -- serial data coming in
        SpiNSS_l     : out std_logic);  -- slave select out
end entity spimaster;

architecture master of spimaster is
    
    signal iMoSiSR     : std_logic_vector(7 downto 0);  -- tx shift register
    signal iMiSoSR     : std_logic_vector(7 downto 0);  -- rx shift register
    signal iShiftCnt   : natural range 0 to 7;          -- count bits shifted
    signal iSpiNSS_l   : std_logic;     -- internal copy of slave select
    signal iSpiNSS_l_d : std_logic;     -- delayed for edge detect
    
begin  -- architecture master

    -- Here's the outgoing shift register.
    -- Basically, always shift, except when the load enable is asserted.
    -- Once the whole word has shifted out, we'll send zeros.
    TxShiftReg : process (SpiClk, Reset_l) is
    begin  -- process TxShiftReg
        if Reset_l = '0' then
            iMoSiSR <= (others => '0');
        elsif falling_edge(SpiClk) then
            if ParDataWr = '1' then
                iMoSiSR <= ParDataOut;
            else
                iMoSiSR <= iMoSiSR(6 downto 0) & '0';
            end if;
        end if;
    end process TxShiftReg;

    -- outgoing serial data bit is always the MSB of the shift register:
    SpiMOSI <= iMoSiSR(7);

    -- Control the slave select.  Assert this when we load the outgoing shift
    -- register with the new word, and kill it when the shift counter goes to
    -- zero.   The shift counter keeps track of how many bits we still need to
    -- shift out.
    SlaveSelect : process (SpiClk, Reset_l) is
    begin  -- process SlaveSelect
        if Reset_l = '0' then
            iSpiNSS_l <= '1';
        elsif falling_edge(SpiClk) then
            if iShiftCnt = 0 then
                -- nothing left to shift:
                iSpiNSS_l <= '1';
            elsif ParDataWr = '1' then
                -- shift has started:
                iSpiNSS_l <= '0';
            end if;
        end if;
    end process SlaveSelect;

    SpiNSS_l <= iSpiNSS_l;

    -- Count the number of bits we've shifted.
    -- This counter is loaded with 7 when a new word is written to the outgoing
    -- shift register.  It then counts down to 0 and stops.
    ShiftCounter : process (SpiClk, Reset_l) is
    begin  -- process ShiftCounter
        if Reset_l = '0' then
            iShiftCnt <= 0;
        elsif falling_edge(SpiClk) then
            if ParDataWr = '1' then
                iShiftCnt <= 7;
            elsif iShiftCnt > 0 then
                iShiftCnt <= iShiftCnt - 1;
            end if;
        end if;
    end process ShiftCounter;

    -- Receive shift register.
    -- Always shifts.
    RxShiftReg : process (SpiClk, Reset_l) is
    begin  -- process RxShiftReg
        if Reset_l = '0' then
            iMiSoSR <= (others => '0');
        elsif rising_edge(SpiClk) then
            iMiSoSR <= iMiSoSR(6 downto 0) & SpiMISO;
        end if;
    end process RxShiftReg;

    -- Received data copied to holding register ParDataIn when slave select
    -- goes away.
    LatchParDataIn : process (SpiClk, Reset_l) is
    begin  -- process LatchParDataIn
        if Reset_l = '0' then
            iSpiNSS_l_d  <= '1';
            ParDataIn    <= (others => '0');
            ParDataInVal <= '0';
        elsif rising_edge(SpiClk) then
            iSpiNSS_l_d <= iSpiNSS_l;

            if (iSpiNSS_l = '1') and (iSpiNSS_l_d = '0') then
                ParDataIn    <= iMiSoSR;
                ParDataInVal <= '1';
            elsif ParDataRd = '1' then
                ParDataInVal <= '0';
            end if;
        end if;
    end process LatchParDataIn;
end architecture master;

