-------------------------------------------------------------------------------
-- Title      : SiLabs C8051F340 EMIF model
-- Project    : 
-------------------------------------------------------------------------------
-- File       : emif_pkg.vhdl
-- Author     : Andy Peters  <devel@latke.net>
-- Company    : ASP Digital
-- Created    : 2006-08-07
-- Last update: 2006-08-07
-------------------------------------------------------------------------------
-- Description: 
-------------------------------------------------------------------------------
-- Copyright (c) 2006 Spectral Instruments, Inc.
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2006-08-07  1.0      andy    Created
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;

package emif_pkg is

    procedure Emif_Init (
        signal EmifA    : out std_logic_vector(15 downto 0);
        signal EmifD    : out std_logic_vector(7 downto 0);
        signal EmifWr_l : out std_logic;
        signal EmifRd_l : out std_logic);

    procedure Emif_Write (
        constant TestAddr  : in  std_logic_vector(15 downto 0);  -- address to write to
        constant TestData  : in  std_logic_vector(7 downto 0);  -- what to write
        signal   SysClk    : in  std_logic;     -- system clock
        signal   EmifA     : out std_logic_vector(15 downto 0);  -- address bus
        signal   EmifD     : out std_logic_vector(7 downto 0);  -- data bus
        signal   EmifWr_l  : out std_logic;     -- write strobe
        constant ADDRSETUP : in  natural := 2;  -- data, addr valid # of clocks before write
        constant WRPW      : in  natural := 2;  -- width of WR\, # of clocks
        constant ADDRHOLD  : in  natural := 2);  -- hold after pulse, # clocks

    procedure Emif_Read (
        constant TestAddr  : in  std_logic_vector(15 downto 0);
        signal   SysClk    : in  std_logic;
        signal   EmifA     : out std_logic_vector(15 downto 0);
        signal   EmifD     : in  std_logic_vector(7 downto 0);
        signal   EmifRd_l  : out std_logic;
        signal   TestData  : out std_logic_vector(7 downto 0);
        constant ADDRSETUP : in  natural := 2;
        constant RDPW      : in  natural := 2;
        constant ADDRHOLD  : in  natural := 2);

end package emif_pkg;

package body emif_pkg is

    -- purpose: Initialize all of the EMIF signals
    procedure Emif_Init (
        signal EmifA    : out std_logic_vector(15 downto 0);
        signal EmifD    : out std_logic_vector(7 downto 0);
        signal EmifWr_l : out std_logic;
        signal EmifRd_l : out std_logic) is
    begin  -- procedure Emif_Init
        EmifA    <= (others => '1');
        EmifD    <= (others => 'Z');
        EmifWr_l <= '1';
        EmifRd_l <= '1';
    end procedure Emif_Init;

    procedure Emif_Write (
        constant TestAddr  : in  std_logic_vector(15 downto 0);  -- address to write to
        constant TestData  : in  std_logic_vector(7 downto 0);  -- what to write
        signal   SysClk    : in  std_logic;     -- system clock
        signal   EmifA     : out std_logic_vector(15 downto 0);  -- address bus
        signal   EmifD     : out std_logic_vector(7 downto 0);  -- data bus
        signal   EmifWr_l  : out std_logic;     -- write strobe
        constant ADDRSETUP : in  natural := 2;  -- data, addr valid # of clocks before write
        constant WRPW      : in  natural := 2;  -- width of WR\, # of clocks
        constant ADDRHOLD  : in  natural := 2) is  -- hold after pulse, # clocks

    begin  -- procedure Emif_Write
        wait until rising_edge(SysClk);
        -- drive address and data:
        EmifA <= TestAddr;
        EmifD <= TestData;

        -- wait the setup time between address, data assertion and strobe
        -- assertion: 
        WaitSetup : for i in 0 to ADDRSETUP-1 loop
            wait until rising_edge(SysClk);
        end loop WaitSetup;  -- i

        -- now, assert the write strobe, and keep it active for the right
        -- number of clocks:
        EmifWr_l <= '0';

        WaitPulseWidth : for j in 0 to WRPW-1 loop
            wait until rising_edge(SysClk);
        end loop WaitPulseWidth;  -- j

        -- clear the strobe:
        EmifWr_l <= '1';

        -- wait hold time before deasserting everything:
        for k in 0 to ADDRHOLD-1 loop
            wait until rising_edge(SysClk);
        end loop;  -- k

        EmifA <= (others => '1');
        EmifD <= (others => 'Z');
    end procedure Emif_Write;
    
    procedure Emif_Read (
        constant TestAddr  : in  std_logic_vector(15 downto 0);
        signal   SysClk    : in  std_logic;
        signal   EmifA     : out std_logic_vector(15 downto 0);
        signal   EmifD     : in  std_logic_vector(7 downto 0);
        signal   EmifRd_l  : out std_logic;
        signal   TestData  : out std_logic_vector(7 downto 0);
        constant ADDRSETUP : in  natural := 2;
        constant RDPW      : in  natural := 2;
        constant ADDRHOLD  : in  natural := 2) is
    begin  -- Emif_Read
        wait until rising_edge(SysClk);
        -- drive address and data:
        EmifA <= TestAddr;

        -- wait the setup time between address, data assertion and strobe
        -- assertion: 
        WaitSetup : for i in 0 to ADDRSETUP-1 loop
            wait until rising_edge(SysClk);
        end loop WaitSetup;  -- i

        -- now, assert the write strobe, and keep it active for the right
        -- number of clocks:
        EmifRd_l <= '0';

        WaitPulseWidth : for j in 0 to RDPW-1 loop
            wait until rising_edge(SysClk);
        end loop WaitPulseWidth;  -- j

        -- grab the data and clear the strobe:
        TestData <= EmifD;
        EmifRd_l <= '1';

        -- wait hold time before deasserting everything:
        for k in 0 to ADDRHOLD-1 loop
            wait until rising_edge(SysClk);
        end loop;  -- k

        EmifA <= (others => '1');
    end procedure Emif_Read;

end package body emif_pkg;
