-------------------------------------------------------------------------------
-- Title      : Testbench for design "glcdtest"
-- Project    : 
-------------------------------------------------------------------------------
-- File       : glcdtest_tb.vhdl
-- Author     : Andy Peters  <devel@latke.net>
-- Company    : ASP Digital
-- Created    : 2006-08-05
-- Last update: 2006-08-12
-- Platform   : ModelSim, XST
-- Standard   : VHDL'93, Math Packages
-------------------------------------------------------------------------------
-- Description: This testbench simulates the EMIF-to-display-buffer mechanism
-- used in the graphic LCD FPGA.  We test this by writing all zeros to the
-- memory, then we do a couple of read-modify-write operations.
-- FirstRMW and SecondRMW are convenient markers that indicate when we're doing
-- the first RMW op and the second.
-------------------------------------------------------------------------------
-- Copyright (c) 2006 ASP Digital
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2006-08-05  1.0      andy    Created
-- 2006-08-07           andy    use emif package and simulate MOVX @DPTR, A and
--                              MOVX A, @DPTR to FPGA memory.
-- 2006-08-11           andy    add generics for EMIF timing control
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.emif_pkg.all;

entity glcdtest_tb is
    generic (
        -- these generics control EMIF timing, in terms of # of SYSCLKs.
        ADDRSETUP : natural := 1;       -- setup before RD/WR asserted
        PW        : natural := 1;       -- width of read or write pulse
        ADDRHOLD  : natural := 2);      -- hold time after RD/WR deassertion
end entity glcdtest_tb;

architecture testbench of glcdtest_tb is

    -- component generics
    constant PIXPERLINE    : integer := 240;
    constant LINESPERFIELD : integer := 64;

    -- component ports
    signal SysClk   : std_logic := '0';
    signal Reset_l  : std_logic := '0';
    signal EmifA    : std_logic_vector(15 downto 0);
    signal EmifD    : std_logic_vector(7 downto 0);
    signal EmifWr_l : std_logic;
    signal EmifRd_l : std_logic;
    signal Ps2Clk   : std_logic;
    signal Ps2Data  : std_logic;
    signal SpiMOSI  : std_logic;
    signal SpiMISO  : std_logic;
    signal SpiClk   : std_logic;
    signal SpiNSS_l : std_logic;
    signal GlcdCl1  : std_logic;
    signal GlcdCL2  : std_logic;
    signal GlcdFLM  : std_logic;
    signal GlcdM    : std_logic;
    signal GlcdD    : std_logic_vector(0 to 3);

    signal TestDataRead : std_logic_vector(7 downto 0);
    signal FirstRMW     : std_logic := '0';
    signal SecondRMW    : std_logic := '0';
    
begin  -- architecture testbench

    SysClk <= not SysClk after 10 NS;

    DUT : entity work.glcdtest
        generic map (
            PIXPERLINE    => PIXPERLINE,
            LINESPERFIELD => LINESPERFIELD)
        port map (
            SysClk   => SysClk,
            Reset_l  => Reset_l,
            EmifA    => EmifA,
            EmifD    => EmifD,
            EmifWr_l => EmifWr_l,
            EmifRd_l => EmifRd_l,
            Ps2Clk   => Ps2Clk,
            Ps2Data  => Ps2Data,
            SpiMOSI  => SpiMOSI,
            SpiMISO  => SpiMISO,
            SpiClk   => SpiClk,
            SpiNSS_l => SpiNSS_l,
            GlcdCl1  => GlcdCl1,
            GlcdCL2  => GlcdCL2,
            GlcdFLM  => GlcdFLM,
            GlcdM    => GlcdM,
            GlcdD    => GlcdD);

    ResetThis : process is
    begin  -- process ResetThis
        Reset_l <= '1' after 444 NS;
        wait;
    end process ResetThis;

    MainTest : process is
        variable maddr_lsb : unsigned(7 downto 0);
        variable sum       : unsigned(7 downto 0);
    begin  -- process MainTest
        if Reset_l = '0' then
            Emif_Init(
                EmifA    => EmifA,
                EmifD    => EmifD,
                EmifWr_l => EmifWr_l,
                EmifRd_l => EmifRd_l);
            wait until Reset_l = '1';
        end if;
        wait for 19 US;

        -- Initialize the "memory" in the FPGA.  do this by simply writing 0 to
        -- it.  Basically, this is a bunch of MOVX @DPTR, A operations.
        wait until rising_edge(SysClk);
        InitMem : for maddr in 4096 to 12287 loop
            Emif_Write(
                TestAddr  => std_logic_vector(to_unsigned(maddr, 16)),
                TestData  => X"00",
                SysClk    => SysClk,
                EmifA     => EmifA,
                EmifD     => EmifD,
                EmifWr_l  => EmifWr_l,
                ADDRSETUP => ADDRSETUP,
                WRPW      => PW,
                ADDRHOLD  => ADDRHOLD);
        end loop InitMem;

        wait for 1 MS;

        -- do read-modify-write.
        -- Use the loop to generate the address and add the 8 LSBs of the
        -- address to the value we read from the memory, and write it back.
        FirstRMW <= '1';
        ReadModifyWrite : for maddr in 4096 to 12287 loop
            Emif_Read (
                TestAddr  => std_logic_vector(to_unsigned(maddr, 16)),
                SysClk    => SysClk,
                EmifA     => EmifA,
                EmifD     => EmifD,
                EmifRd_l  => EmifRd_l,
                TestData  => TestDataRead,
                ADDRSETUP => ADDRSETUP,
                RDPW      => PW,
                ADDRHOLD  => ADDRHOLD);

            maddr_lsb := to_unsigned(maddr, 8);
            sum       := maddr_lsb + unsigned(TestDataRead);
            
            Emif_Write (
                TestAddr  => std_logic_vector(to_unsigned(maddr, 16)),  -- address to write to
                TestData  => std_logic_vector(sum),  -- what to write
                SysClk    => SysClk,    -- system clock
                EmifA     => EmifA,     -- address bus
                EmifD     => EmifD,     -- data bus
                EmifWr_l  => EmifWr_l,  -- write strobe
                ADDRSETUP => ADDRSETUP,
                WRPW      => PW,
                ADDRHOLD  => ADDRHOLD);

            wait for 1 US;
        end loop ReadModifyWrite;
        FirstRMW <= '0';

        wait for 1 MS;

        -- do read-modify-write.
        -- Use the loop to generate the address and add the 8 LSBs of the
        -- address to the value we read from the memory, and write it back.
        SecondRMW <= '1';
        ReadModifyWriteAgain : for maddr in 4096 to 12287 loop
            Emif_Read (
                TestAddr  => std_logic_vector(to_unsigned(maddr, 16)),
                SysClk    => SysClk,
                EmifA     => EmifA,
                EmifD     => EmifD,
                EmifRd_l  => EmifRd_l,
                TestData  => TestDataRead,
                ADDRSETUP => ADDRSETUP,
                RDPW      => PW,
                ADDRHOLD  => ADDRHOLD);

            maddr_lsb := to_unsigned(maddr, 8);
            sum       := maddr_lsb + unsigned(TestDataRead);
            
            Emif_Write (
                TestAddr  => std_logic_vector(to_unsigned(maddr, 16)),  -- address to write to
                TestData  => std_logic_vector(sum),  -- what to write
                SysClk    => SysClk,    -- system clock
                EmifA     => EmifA,     -- address bus
                EmifD     => EmifD,     -- data bus
                EmifWr_l  => EmifWr_l,  -- write strobe
                ADDRSETUP => ADDRSETUP,
                WRPW      => PW,
                ADDRHOLD  => ADDRHOLD);

            wait for 1 US;
        end loop ReadModifyWriteAgain;
        SecondRMW <= '0';
        assert (FALSE) report "Done!" severity FAILURE;
        wait;
    end process MainTest;

end architecture testbench;
